unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, RSQ, RSRADPlay, RSGraphics, RSSysUtils, Menus, RSTimer;

{
[+] Now the player can be controlled with mouse (wheel & double click)
[-] Pressing Alt+Enter while holding a mouse key worked incorrectly
[-] Interlaced videos weren't shown properly
}

type
  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    PopupMenu1: TPopupMenu;
    FullScreen1: TMenuItem;
    Timer1: TRSTimer;
    Timer2: TRSTimer;
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Timer1Timer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FullScreen1Click(Sender: TObject);
    procedure FormContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure Timer2Timer(Sender: TObject);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WMActivate(var Msg:TWMActivate); message WM_Activate;
  public
    function LoadFile(f: string = ''):Boolean;
    procedure Reposition;
    procedure RefreshCaption(const Middle:string); overload;
    procedure RefreshCaption(ShowCadre:Boolean); overload;
    procedure RefreshCaption; overload;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

const
  FormTitle = 'RS Bink Player';
  HalfToneText: array[Boolean] of string =
    ('Simple Stretching', 'Halftone Stretching On');

var
  BinkNew, BinkOld:TRSBinkPlayer; Smack:TRSSmackPlayer;
  Player: TRSRADPlayer;
  Picture: TBitmap;
  Working, Stopped, FullScreen, ShowCadre: Boolean;
  TimerEnabled: Boolean;
  FileName, LastFile: string;
  BlockActivate, DeactPause: Boolean;
  NeedErase, HalfToneStretch, HalftoneSwitched: Boolean;

procedure TForm1.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.WinClassName:= 'RS Bink Player Main Form';
end;

procedure TForm1.WMActivate(var Msg: TWMActivate);
begin
  inherited;
  if FullScreen then
    if Msg.Active<>WA_INACTIVE then
    begin
      Application.Restore;
      Show;
      Perform(WM_SYSCOMMAND, SC_MAXIMIZE, 0);
      ShowCursor(false);
    end else
    begin
      Application.Minimize;
      Hide;
    end;

  if not DeactPause and not FullScreen then  exit;
  Timer1.Enabled:= TimerEnabled and (Msg.Active<>WA_INACTIVE);
  if (Player<>nil) and not BlockActivate then
    Player.Pause:= not Timer1.Enabled;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FreeAndNil(Player);
  Stopped:=true;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  if Player.Wait then  exit;
  if not FullScreen then
  begin
    Player.NextFrame;
    Player.ExtractFrame(Picture);
  end else
    repeat
      Player.NextFrame;
      Player.ExtractFrame(Picture);
    until Player.Wait;
    
  if ShowCadre then
    RefreshCaption;
  InvalidateRect(Handle, nil, false);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  i,j,k:int; s,s1,s2:string; SaveFrame:int; b:TBitmap;
begin
  if (Key<>VK_ESCAPE) and Working then
    exit;

  case Key of
    VK_ESCAPE:
      begin
        Stopped:=true;
        if not Working then
          Close;
      end;

    ord('O'), ord('L'):
      LoadFile;

    VK_SPACE:
    begin
      TimerEnabled:= not TimerEnabled and (Player.FramesCount>1);
      Player.Pause:= not TimerEnabled;
      Timer1.Enabled:= TimerEnabled;
    end;

    VK_RIGHT, VK_LEFT:
    begin
      Player.Pause:=true;
      Timer1.Enabled:=false;
      if Key = VK_LEFT then
      begin
        RefreshCaption(' (wait) - ');
        Player.GotoFrame(Player.Frame - 1, Picture);
      end else
        Player.GotoFrame(Player.Frame + 1, Picture);
      RefreshCaption;
      InvalidateRect(Handle, nil, false);
    end;

    VK_PGUP, VK_PGDN, VK_HOME, VK_END:
    begin
      if FileName = '' then
        exit;

      with TRSFindFile.Create(FileName) do
      begin
        if Found then
          s2:=Data.cFileName
        else
          s2:=ExtractFileName(Unit1.FileName);
        Free;
      end;

      with TRSFindFile.Create(ExtractFilePath(FileName)+'*.*') do
        try
          while FindAttributes(0, FILE_ATTRIBUTE_DIRECTORY) do // Only files
          begin
            s:=ExtractFileExt(Data.cFileName);
            if SameText(s,'.bik') or SameText(s,'.smk') then
            begin
              s:=Data.cFileName;

              case Key of
                VK_PGUP:
                  if SameText(s, s2) then
                    break;
                VK_PGDN:
                  if SameText(s1, s2) then
                    break;
                VK_HOME:
                  break;
              end;
              s1:=s;
            end;
            s:='';
            FindNext;
          end;
        finally
          Free;
        end;

      if (s<>'') or (Key = VK_END) and (s1<>'') then
      begin
        if Key in [VK_PGUP, VK_END] then
          s:=s1;
        if s<>'' then
          LoadFile(ExtractFilePath(FileName)+s);
      end;
    end;
    
    ord('Z'):
      if (ssCtrl in Shift) and not (ssAlt in Shift) and FileExists(LastFile) then
        LoadFile(LastFile);

    ord('V'):
    begin
      ShowCadre:= not ShowCadre;
      RefreshCaption;
    end;

    ord('H'):
    begin
      HalfToneStretch:= not HalfToneStretch;
      NeedErase:= NeedErase or HalftoneSwitched;
      HalftoneSwitched:= true;
      InvalidateRect(Handle, nil, false);
      Timer2.Enabled:= false;
      Timer2.Enabled:= true;
    end;

    ord('D'):
      DeactPause:= not DeactPause;

    ord('S'):
    begin
      j:= Player.FramesCount;
      k:= length(IntToStr(j-1));
      if TimerEnabled then
        s:= ChangeFileExt(ExtractFileName(OpenDialog1.FileName), '') +
             '\Extract Here'
      else
        s:= Format('%.*d.bmp', [k, Player.Frame]);

      SaveDialog1.FileName:=s;
      if not SaveDialog1.Execute then  exit;
      if TimerEnabled then
      begin
        s:= ExtractFilePath(SaveDialog1.FileName);
        RSCreateDir(s);
        SaveFrame:= Player.Frame;
        b:= TBitmap.Create;
        b.Assign(Picture);
        Timer1.Enabled:= false;
        Player.Pause:= true;
        Working:= true;
        Stopped:= false;
        Application.ProcessMessages;
        try
          for i:=1 to j do  //  j = Player.FramesCount
          begin
            if Stopped then
              break;

            b.SaveToFile(s + Format('%.*d.bmp', [k, Player.Frame]));
            RefreshCaption(true);
            Player.Frame:=Player.Frame+1;
            if i<>j then
              Player.ExtractFrame(b);
            Application.ProcessMessages;
          end;
        finally
          Working:=false;
          RefreshCaption;
          b.Free;
          if Player<>nil then
          begin
            Player.Frame:=SaveFrame;
            Timer1.Enabled:=TimerEnabled;
            Player.Pause:= not TimerEnabled;
          end;
        end;
      end else
      begin
        s:=SaveDialog1.FileName;
        RSCreateDir(ExtractFilePath(s));
        Picture.SaveToFile(s);
      end;
    end;
  end;
end;

function TForm1.LoadFile(f: string = ''):Boolean;
const
  SMK='SMK';
var
  hf:HFile; s:string; i:int; Play:TRSRADPlayer; Changed:Boolean;
begin
  Result:=false;
  if (f<>'') and SameText(f, FileName) then  exit;
  Working:=true;
  Timer1.Enabled:=false;
  if Player<>nil then
    Player.Pause:=true;
  try
    OpenDialog1.InitialDir:=AppPath;
    OpenDialog1.FileName:=f;
    if f='' then
    begin
      BlockActivate:=true;
      if OpenDialog1.Execute then
        f:=OpenDialog1.FileName;
      Application.ProcessMessages;
      BlockActivate:=false;
    end;

    Result:=f<>'';
    if not Result then
      exit;

    s:=ExtractFileExt(f);
    if not SameText(s, '.smk') then
    begin
      hf:= CreateFile(ptr(f), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING,
             FILE_ATTRIBUTE_NORMAL, 0);
      if hf = INVALID_HANDLE_VALUE then
        RaiseLastOSError;
      i:=0;
      FileRead(hf, i, 4);
      if i and $ffffff <> pint(PChar(SMK))^ then
      begin
        i:=i shr 24;
        if i>$62 then
          Play:=BinkNew
        else
          Play:=BinkOld;
      end else
        Play:=Smack;
      CloseHandle(hf);
    end else
      Play:=Smack;

    if Player<>nil then
      Player.Close;

    Play.Open(f);

    if Player<>nil then
      Sleep(1); // Avoid Bink and Smack thread synchronization bug in most cases
                // RSSetSilentExceptionsFiter is used to handle other cases

    Changed:=(Picture.Width <> Play.Width^) or (Picture.Height <> Play.Height^);
    if Changed or (Player <> Play) then
      Picture.Assign(nil);
    with Picture.Canvas do
    begin
      Brush.Color:= clBlack;
      FillRect(ClipRect); // Important for interlaced videos
    end;

    Play.PreparePic(Picture);
    Play.ExtractFrame(Picture);
    Player:=Play;
    LastFile:=FileName;
    FileName:=f;
    RefreshCaption;
    NeedErase:= NeedErase or Changed;
    InvalidateRect(Handle, nil, false);
    Reposition;
    UpdateWindow(Handle);
//    Application.ProcessMessages;  // Slower
    TimerEnabled:=Player.FramesCount>1;
  finally
    Timer1.Enabled:=TimerEnabled;
    if Player<>nil then
      Player.Pause:=not TimerEnabled;
    Working:=false;
    Stopped:=false;  
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  RSSetSilentExceptionsFiter;
  Picture:=TBitmap.Create;
  BinkNew:=TRSBinkPlayer.Create('BinkW32New.dll');
  if BinkNew.LibInstance=0 then
  begin
    BinkNew.Free;
    BinkNew:=TRSBinkPlayer.Create;
  end;
  BinkOld:=TRSBinkPlayer.Create('BinkW32Old.dll');
  if BinkOld.LibInstance=0 then
  begin
    BinkOld.Free;
    BinkOld:=TRSBinkPlayer.Create;
  end;

  Smack:=TRSSmackPlayer.Create;
  try
    Application.ShowMainForm:= LoadFile(ParamStr(1));
    if Application.ShowMainForm then
    begin
      BlockActivate:=true;
      SetForegroundWindow(Handle);
      SetActiveWindow(Application.Handle);
      BlockActivate:=false;
    end else
      PostMessage(Handle, WM_CLOSE, 0, 0);
  except
    on e:Exception do
    begin
      SetForegroundWindow(Handle);
      SetActiveWindow(Application.Handle);
      Application.ShowMainForm:=false;
      Application.ShowException(e);
      PostMessage(Handle, WM_CLOSE, 0, 0);
    end;
  end;
end;

procedure TForm1.FormPaint(Sender: TObject);
var h:int; r, r1:TRect;
begin
  if (Picture.Width = 0) or (Picture.Height = 0) then  exit;
  
  with Canvas do
    if FullScreen then
    begin
      if HalfToneStretch then
        SetStretchBltMode(Handle, HALFTONE)
      else
        SetStretchBltMode(Handle, BLACKONWHITE);

      if Picture.Width*Picture.Height <> 0 then
      begin
        //r:= ClientRect;
        r:= Rect(0, 0, Screen.Width, Screen.Height);
        h:= (r.Right - r.Left)*Picture.Height div Picture.Width;
        if h > r.Bottom - r.Top then
        begin
          h:= (r.Bottom - r.Top)*Picture.Width div Picture.Height;
          r.Left:= (r.Right - h) div 2;
          r.Right:= r.Left + h;
        end else
        begin
          r.Top:= (r.Bottom - h) div 2;
          r.Bottom:= r.Top + h;
        end;
        CopyRect(r, Picture.Canvas, Picture.Canvas.ClipRect);
      end;
      if NeedErase then
      begin
        Brush.Color:= Color;
        Brush.Style:= bsSolid;
        r1:= ClipRect;
        if r.Top > r1.Top then
          FillRect(Rect(r1.Left, r1.Top, r1.Right, r.Top));
        if r.Bottom < r1.Bottom then
          FillRect(Rect(r1.Left, r.Bottom, r1.Right, r1.Bottom));
        if r.Left > r1.Left then
          FillRect(Rect(r1.Left, r1.Top, r.Left, r1.Bottom));
        if r.Right < r1.Right then
          FillRect(Rect(r.Right, r1.Top, r1.Right, r1.Bottom));
      end;
    end else
      CopyRect(ClipRect, Picture.Canvas, ClipRect);

  NeedErase:=false;

  if HalftoneSwitched then
    with Canvas do
    begin
      Font:= self.Font;
      Font.Size:= 12;
      Brush.Style:= bsClear;
      Font.Color:= clBlue;
      TextOut(11, 11, HalftoneText[HalfToneStretch]);
      Font.Color:= clAqua;
      TextOut(10, 10, HalftoneText[HalfToneStretch]);
    end;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key in [VK_LEFT, VK_RIGHT] then
  begin
    Timer1.Enabled:= TimerEnabled;
    Player.Pause:= not TimerEnabled;
  end;
end;

procedure TForm1.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var
  Key:Word;
begin
  if WheelDelta > 0 then
    Key:= VK_PGUP
  else
    Key:= VK_PGDN;
  FormKeyDown(self, Key, []);
end;

procedure TForm1.FullScreen1Click(Sender: TObject);
const
  AStyle = int($97000000 xor $16CA0000 xor WS_MAXIMIZE);
  AExStyle = $10000 xor $10100;
begin
  ReleaseCapture;
  FullScreen:= not FullScreen;
  with TRSWnd(Handle) do
  begin
    Style:= Style xor AStyle;
    ExStyle:= ExStyle xor AExStyle;
  end;

  if FullScreen then
    Perform(WM_SYSCOMMAND, SC_MAXIMIZE, 0)
  else
    Perform(WM_SYSCOMMAND, SC_RESTORE, 0);

  if not FullScreen then
  begin
    SetWindowPos(Handle, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE or SWP_FRAMECHANGED);
    Reposition;
  end else
    SetWindowPos(Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE or SWP_FRAMECHANGED);

  ShowCursor(not FullScreen);
  Invalidate;
end;

procedure TForm1.FormContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  Handled:=true;
end;

procedure TForm1.Reposition;
var w,h:int;
begin
  if FullScreen then  exit;

  w:= Picture.Width + Width - ClientWidth;
  h:= Picture.Height + Height - ClientHeight;

  SetBounds(Monitor.Left + (Monitor.Width - w) div 2,
            Monitor.Top + (Monitor.Height - h) div 2,
            w, h);
end;

procedure TForm1.RefreshCaption(const Middle: string);
begin
  Caption:= ExtractFileName(FileName) + Middle + FormTitle;
end;

procedure TForm1.RefreshCaption(ShowCadre:Boolean);
begin
  if ShowCadre then
    RefreshCaption(Format(' (%d/%d) - ', [Player.Frame, Player.FramesCount-1]))
  else
    RefreshCaption(' - ');
end;

procedure TForm1.RefreshCaption;
begin
  RefreshCaption(ShowCadre);
end;

procedure TForm1.Timer2Timer(Sender: TObject);
begin
  HalftoneSwitched:= false;
  Timer2.Enabled:= false;
  NeedErase:= true;
  InvalidateRect(Handle, nil, false);
end;

end.
